import os
import os.path as op
import logging as log
from tqdm import tqdm


def download(x, experiments, resource_name, validator, destdir, subcommand):
    """Collect resources from a given set of experiments, given a resource
    name, a validator name, a destination folder and a subcommand.

    Examples of subcommand are: files, report, snapshot, rc, layers/lobes,
    maps"""

    if len(experiments) > 1:
        log.warning('Now initiating download for %s experiments.'
                    % len(experiments))
        experiments = tqdm(experiments)
    for e in experiments:
        log.debug(e)
        try:
            kwargs = {'session_id': e['ID'],
                      'subject_label': e['subject_label'],
                      'validator': validator,
                      'debug': False}
            pattern = '%s_' % subcommand.capitalize()
            kwargs['pattern'] = pattern

            subcommands = {'files': __dl_files__,
                           'report': __dl_report__,
                           'snapshot': __dl_snap__,
                           'rc': __dl_rc__,
                           'maps': __dl_maps__,
                           'layers': __dl_bamos__,
                           'lobes': __dl_bamos__}

            if subcommand in subcommands.keys():
                fp = __download__(x, e, resource_name, destdir,
                                  subcommands[subcommand], **kwargs)
            else:
                raise Exception('Invalid subcommand (%s).' % subcommand)

        except KeyboardInterrupt:
            return

    if len(experiments) == 1 and fp is not None:
        log.info('Saving it in %s' % fp)


def __download__(x, e, resource_name, destdir, func, **kwargs):
    debug = kwargs.get('debug', False)
    e_id = e['ID']
    log.debug(e_id)
    e = x.select.experiment(e_id)
    r = e.resource(resource_name)
    if not r.exists():
        log.error('%s has no %s resource' % (e_id, resource_name))
        return
    if debug:
        fp = func(e, r, destdir, **kwargs)
    else:
        try:
            fp = func(e, r, destdir, **kwargs)
        except Exception as exc:
            log.error('%s failed. Skipping (%s).' % (e_id, exc))
            return
    return fp


def __dl_report__(e, r, destdir, **kwargs):
    v = e.resource('BBRC_VALIDATOR')
    f = v.pdf(kwargs['validator'])
    fp = op.join(destdir, f.label())
    f.get(dest=fp)
    return fp


def __dl_files__(e, r, destdir, **kwargs):
    e_id = kwargs['session_id']
    validator = kwargs['validator']

    dd = op.join(destdir, e_id)
    if op.isdir(dd):
        msg = '%s (%s) already exists. Skipping folder creation.' % (dd, e_id)
        log.error(msg)
    else:
        os.mkdir(dd)
    r.get(dest_dir=dd)

    kwargs['validator'] = validator
    __dl_report__(e, r, dd, **kwargs)

    return destdir


def __dl_maps__(e, r, destdir, **kwargs):
    r.download_maps(destdir)
    return destdir


def __dl_snap__(e, r, destdir, **kwargs):
    e_id = kwargs['session_id']
    r = e.resource('BBRC_VALIDATOR')
    fp = op.join(destdir, '%s.jpg' % e_id)
    fp = r.download_snapshot(kwargs['validator'], fp)
    return ', '.join(fp)


def __dl_rc__(e, r, destdir, **kwargs):
    subject_label = kwargs['subject_label']
    e_id = kwargs['session_id']
    validator = kwargs['validator']

    r.download_rc(destdir)
    v = e.resource('BBRC_VALIDATOR')
    if v.exists():
        fp = op.join(destdir, '%s_%s.jpg' % (subject_label, e_id))
        try:
            v.download_snapshot(validator, fp)
        except Exception as exc:
            log.error('%s has no %s (%s)' % (e_id, validator, exc))
    else:
        log.warning('%s has not %s' % (e, validator))
    return destdir


def __dl_bamos__(e, r, destdir, **kwargs):
    pattern = kwargs['pattern']
    subject_label = kwargs['subject_label']
    e_id = kwargs['session_id']
    f = list(r.files('%s*' % pattern))[0]
    pattern = pattern.lower().rstrip('_')
    fp = op.join(destdir,
                 '%s_%s_%s.nii.gz' % (subject_label, e_id, pattern))
    f.get(fp)
    return fp


def __fix_volumes__(volumes):
    """Remove incorrect volumes by FREESURFER6_HIRES as mentioned in the
    following page.

    Reference: https://surfer.nmr.mgh.harvard.edu/fswiki/BrainVolStatsFixed
    """
    metrics = ['BrainSegVol', 'BrainSegVolNotVent', 'SupraTentorialVol',
               'lhCerebralWhiteMatterVol', 'rhCerebralWhiteMatterVol',
               'CerebralWhiteMatterVol', 'TotalGrayVol', 'SubCortGrayVol',
               'SupraTentorialVolNotVent', 'MaskVol', 'MaskVol-to-eTIV',
               'BrainSegVol-to-eTIV', 'lhCortexVol', 'rhCortexVol',
               'CortexVol']
    volumes = volumes.drop(volumes.query('region.isin(@metrics)').index)
    return volumes


def __braak_fdg__(x, e_id, r):
    import bx
    import numpy as np
    import pandas as pd
    import tempfile
    import nibabel as nib
    resource_name = 'FDG_QUANTIFICATION2'
    fh, fp = tempfile.mkstemp(suffix='.nii.gz')
    os.close(fh)
    r = x.select.experiment(e_id).resource(resource_name)
    f = r.file('GLOBAL/wstatic_pet_scaled_vermis.nii.gz')
    f.get(fp)

    regions_dir = op.join(op.dirname(bx.__file__), 'data', 'braak')

    columns = ['Braak_I_II', 'Braak_III_IV', 'Braak_V_VI']
    df = pd.DataFrame(index=[e_id], columns=columns)
    for region in columns:
        atlas_fp = op.join(regions_dir, '%s.nii.gz' % region)
        atlas_im = nib.load(atlas_fp)
        atlas = np.array(atlas_im.dataobj)
        m = np.array(nib.load(fp).dataobj)
        assert (m.shape == atlas.shape)
        n_labels = list(np.unique(atlas))
        res = {label: np.mean(m[atlas == label]) for label in n_labels}
        df[region] = res[n_labels[1]]
    os.remove(fp)
    return df


def __perfusion__(x, e_id, r):
    import bx
    import numpy as np
    import pandas as pd
    import tempfile
    import nibabel as nib
    resource_name = 'BASIL'
    fh, fp = tempfile.mkstemp(suffix='.nii.gz')
    os.close(fh)
    r = x.select.experiment(e_id).resource(resource_name)
    f = r.file('/std_space/perfusion_calib.nii.gz')
    f.get(fp)
    regions = ['Background', 'Precentral_L', 'Precentral_R', 'Frontal_Sup_2_L',
               'Frontal_Sup_2_R', 'Frontal_Mid_2_L', 'Frontal_Mid_2_R',
               'Frontal_Inf_Oper_L', 'Frontal_Inf_Oper_R', 'Frontal_Inf_Tri_L',
               'Frontal_Inf_Tri_R', 'Frontal_Inf_Orb_2_L',
               'Frontal_Inf_Orb_2_R', 'Rolandic_Oper_L', 'Rolandic_Oper_R',
               'Supp_Motor_Area_L',	'Supp_Motor_Area_R', 'Olfactory_L',
               'Olfactory_R', 'Frontal_Sup_Medial_L', 'Frontal_Sup_Medial_R',
               'Frontal_Med_Orb_L',	'Frontal_Med_Orb_R', 'Rectus_L',
               'Rectus_R', 'OFCmed_L', 'OFCmed_R', 'OFCant_L', 'OFCant_R',
               'OFCpost_L',	'OFCpost_R', 'OFClat_L', 'OFClat_R', 'Insula_L',
               'Insula_R', 'Cingulate_Ant_L', 'Cingulate_Ant_R',
               'Cingulate_Mid_L', 'Cingulate_Mid_R', 'Cingulate_Post_L',
               'Cingulate_Post_R', 'Hippocampus_L', 'Hippocampus_R',
               'ParaHippocampal_L', 'ParaHippocampal_R', 'Amygdala_L',
               'Amygdala_R', 'Calcarine_L', 'Calcarine_R', 'Cuneus_L',
               'Cuneus_R', 'Lingual_L', 'Lingual_R', 'Occipital_Sup_L',
               'Occipital_Sup_R', 'Occipital_Mid_L', 'Occipital_Mid_R',
               'Occipital_Inf_L', 'Occipital_Inf_R', 'Fusiform_L',
               'Fusiform_R', 'Postcentral_L', 'Postcentral_R',
               'Parietal_Sup_L', 'Parietal_Sup_R', 'Parietal_Inf_L',
               'Parietal_Inf_R', 'SupraMarginal_L',	'SupraMarginal_R',
               'Angular_L',	'Angular_R', 'Precuneus_L',	'Precuneus_R',
               'Paracentral_Lobule_L', 'Paracentral_Lobule_R',	'Caudate_L',
               'Caudate_R', 'Putamen_L', 'Putamen_R', 'Pallidum_L',
               'Pallidum_R', 'Thalamus_L', 'Thalamus_R', 'Heschl_L',
               'Heschl_R', 'Temporal_Sup_L', 'Temporal_Sup_R',
               'Temporal_Pole_Sup_L', 'Temporal_Pole_Sup_R', 'Temporal_Mid_L',
               'Temporal_Mid_R', 'Temporal_Pole_Mid_L',	'Temporal_Pole_Mid_R',
               'Temporal_Inf_L', 'Temporal_Inf_R', 'Cerebelum_Crus1_L',
               'Cerebelum_Crus1_R', 'Cerebelum_Crus2_L', 'Cerebelum_Crus2_R',
               'Cerebelum_3_L', 'Cerebelum_3_R', 'Cerebelum_4_5_L',
               'Cerebelum_4_5_R', 'Cerebelum_6_L', 'Cerebelum_6_R',
               'Cerebelum_7b_L', 'Cerebelum_7b_R', 'Cerebelum_8_L',
               'Cerebelum_8_R', 'Cerebelum_9_L', 'Cerebelum_9_R',
               'Cerebelum_10_L', 'Cerebelum_10_R', 'Vermis_1_2', 'Vermis_3',
               'Vermis_4_5', 'Vermis_6', 'Vermis_7', 'Vermis_8', 'Vermis_9',
               'Vermis_10']

    df = pd.DataFrame()
    atlas_fp = op.join(op.dirname(bx.__file__), 'data', 'aal2', 'AAL2.nii')
    atlas_im = nib.load(atlas_fp)
    atlas = np.array(atlas_im.dataobj)
    m = np.array(nib.load(fp).dataobj)
    assert (m.shape == atlas.shape)
    n_labels = list(np.unique(atlas))
    df['region'] = regions
    df['ID'] = [e_id] * len(n_labels)
    perf_values = []
    for label in n_labels:
        perf_values.append(np.mean(m[atlas == label]))
    df['value'] = perf_values
    os.remove(fp)
    return df


def __aging_fdg__(x, e_id, r):
    import bx
    import numpy as np
    import pandas as pd
    import tempfile
    import nibabel as nib
    resource_name = 'FDG_QUANTIFICATION2'
    fh, fp = tempfile.mkstemp(suffix='.nii.gz')
    os.close(fh)
    r = x.select.experiment(e_id).resource(resource_name)
    f = r.file('GLOBAL/wstatic_pet_scaled_pons.nii.gz')
    f.get(fp)

    cols = ['optimized_pet', 'region', 'reference_region', 'measurement', 'value']
    row = [False, 'aging_fdg_composite', 'pons', 'suvr', None]
    df = pd.DataFrame(data=[row], index=[e_id], columns=cols)
    aging_regions = op.join(op.dirname(bx.__file__), 'data',
                            'masks', 'fdg_aging_mask.nii')
    ag_im = nib.load(aging_regions)
    ag_regs = np.array(ag_im.dataobj)
    m = np.array(nib.load(fp).dataobj)
    assert (m.shape == ag_regs.shape)
    df['value'] = np.mean(m[ag_regs == 1])
    os.remove(fp)
    return df


def measurements(x, experiments, subfunc, resource_name='FREESURFER6',
                 debug=False):
    """ Collect measurements for a set of experiments by calling some specific
    pyxnat resource-based function (e.g. aseg, aparc, centiloids, etc)"""
    import pandas as pd

    table = []
    for e in tqdm(experiments):
        log.debug(e)
        try:
            s = e['subject_label']
            e_id = e['ID']
            r = x.select.experiment(e_id).resource(resource_name)
            if not r.exists():
                log.error('%s has no %s resource' % (e_id, resource_name))
                continue

            if subfunc == 'aparc':
                volumes = r.aparc()
                if resource_name.endswith('_HIRES'):
                    volumes = __fix_volumes__(volumes)
            elif subfunc == 'aseg':
                volumes = r.aseg()
                if resource_name.endswith('_HIRES'):
                    volumes = __fix_volumes__(volumes)
            elif subfunc == 'centiloids':
                cl = r.centiloids(optimized=False)
                volumes = pd.DataFrame([cl], columns=[subfunc])
            elif subfunc == 'landau':
                landau = r.landau_signature(optimized=False)
                volumes = pd.DataFrame([landau], columns=[subfunc])
            elif subfunc == 'pet_hammers':
                volumes = r.regional_quantification(optimized=True, atlas='hammers')
            elif subfunc == 'pet_aal':
                volumes = r.regional_quantification(optimized=True, atlas='aal')
            elif subfunc == 'hippoSfVolumes':
                volumes = r.hippoSfVolumes(mode='T1')
            elif subfunc == 'amygNucVolumes':
                volumes = r.amygNucVolumes()
            elif subfunc == 'bamos_volumes':
                volumes = pd.DataFrame([r.volume()], columns=['volume'])
            elif subfunc == 'bamos_stats':
                volumes = r.stats()
            elif subfunc == 'volumes':
                volumes = r.volumes()
                if isinstance(volumes, dict):
                    volumes = pd.DataFrame(volumes, index=[s])
            elif subfunc == 'fdg':
                volumes = __braak_fdg__(x, e_id, r)
            elif subfunc == 'basil_perfusion':
                volumes = r.perfusion()
            elif subfunc in ['basil_stats', 'qsmxt_stats']:
                volumes = r.stats()
            elif subfunc == 'basil_aal':
                volumes = __perfusion__(x, e_id, r)
            elif subfunc == 'asl3d_perfusion':
                volumes = r.perfusion().query('atlas  == "global"')
            elif subfunc == 'asl3d_aal':
                volumes = r.perfusion().query('atlas  == "aal2"')
            elif subfunc == 'aging':
                volumes = __aging_fdg__(x, e_id, r)
            elif subfunc in ['brainstem', 'thalamus', 'hypothalamus']:
                cmd = {'brainstem': 'brainstem_substructures_volumes',
                       'thalamus': 'thalamic_nuclei_volumes',
                       'hypothalamus': 'hypothalamic_subunits_volumes'}
                volumes = eval('r.{}()'.format(cmd[subfunc]))
            volumes['subject'] = s
            volumes['ID'] = e['ID']
            table.append(volumes)
        except KeyboardInterrupt:
            return pd.concat(table).set_index('ID').sort_index()
        except Exception as exc:
            if debug:
                raise exc
            else:
                log.error('Failed for %s. Skipping it. (%s)' % (e, exc))

    data = pd.concat(table).set_index('ID').sort_index()
    return data



import os.path as op
import os
import tempfile
import argparse
import pkgutil
import inspect
import logging as log
from rich.console import Console
from rich.theme import Theme
t = Theme()
t.styles.pop('repr.number')
t.styles.pop('repr.str')
console = Console(theme=t)
print = console.print


def __get_modules__(m):
    modules = []
    prefix = m.__name__ + '.'
    log.info('prefix : %s' % prefix)
    for importer, modname, ispkg in pkgutil.iter_modules(m.__path__, prefix):
        module = __import__(modname, fromlist='dummy')
        if not ispkg:
            modules.append(module)
        else:
            modules.extend(__get_modules__(module))
    return modules


def __find_all_commands__(m):
    """Browses bx and looks for any class named as a Command"""
    modules = []
    classes = []
    modules = __get_modules__(m)
    forbidden_classes = []  # Test, ScanTest, ExperimentTest]
    for m in modules:
        for name, obj in inspect.getmembers(m):
            if inspect.isclass(obj) and 'Command' in name \
                    and obj not in forbidden_classes:
                classes.append(obj)
    return classes


class ReadableDir(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        prospective_dir = values
        if not os.path.isdir(prospective_dir):
            msg = "readable_dir:{0} is not a valid path".format(prospective_dir)
            raise argparse.ArgumentTypeError(msg)
        if os.access(prospective_dir, os.R_OK):
            setattr(namespace, self.dest, prospective_dir)
        else:
            msg = "readable_dir:{0} is not a readable dir".format(prospective_dir)
            raise argparse.ArgumentTypeError(msg)


def __stylize_help__(doc):
    ln = doc.split('\n')
    sep = ':\t'
    subcommands = [e.split(sep)[0].strip(' ') for e in ln if sep in e]
    msg = '[italic]%s[/italic]\n%s' % (ln[0], '\n'.join(ln[1:]))
    for e in subcommands:
        msg = msg.replace(' %s:' % e, ' [bold cyan]%s[/bold cyan]:' % e)
    i = ln.index([e for e in ln if 'Usage:' in e][0])
    msg = msg.replace(ln[i+1],
                      '[bright_white on black]%s[/bright_white on black]' % ln[i+1])
    return msg


def parse(parser, command):
    c = command(parser.command, parser.args, parser.xnat, parser.destdir)

    if len(c.args) == 0:
        msg = __stylize_help__(c.__doc__)
        cn = parser.command
        msg = '\n[bold]Help for command `[cyan]%s[/cyan]`:[/bold]'\
              '\n\n%s' % (cn, msg)
        print(msg)

    elif len(c.args) == command.nargs:
        if command.nargs == 1 or c.args[0] in command.subcommands:
            c.parse()
        else:
            subcommands = ['[bold cyan]%s[/bold cyan]' % e
                           for e in command.subcommands]
            subcommands = '\n - '.join(subcommands)
            args = (c.args[0], subcommands)
            msg = '\n`[yellow]%s[/yellow]` is an invalid subcommand.\n\nAvailable subcommands:\n - %s' % args
            print(msg)

    else:
        msg = '[yellow]%s[/yellow]' % c.__doc__
        cn = parser.command
        args = (cn, msg)
        msg = '\n[bold][red]ERROR[/red]: Missing argument(s)\n\nHelp for command `[blue]%s[/blue]`:[/bold]\n\n%s' % args
        print(msg)


def parse_args(command, args, x, destdir=tempfile.gettempdir()):
    from bx.command import Command
    parser = Command(command, args, x, destdir)

    import bx
    commands = __find_all_commands__(bx)
    commands = {e.__name__.split('.')[-1].lower()[:-7]: e for e in commands}

    if command in commands.keys():
        autorun = any([os.environ.get('CI_TEST', None),
                       os.environ.get('BX_DUMP', None)])

        print('Command: %s' % command)
        if command == 'nifti' and len(args) == 1:
            args.insert(0, 'usable')
        elif command in ['freesurfer6', 'freesurfer']:
            ans = 1
            msg = 'Please confirm if you want FREESURFER6, ' \
                  'FREESURFER6_HIRES or FREESURFER7? '\
                  ' (1) FREESURFER6 (2) FREESURFER6_HIRES (3) FREESURFER7 ?'
            if not autorun:
                while ans not in ['1', '2', '3']:
                    ans = input(msg)
            options = ['freesurfer6', 'freesurfer6hires', 'freesurfer7']
            command = options[int(ans) - 1]
            parser = Command(command, args, x, destdir)
        elif command in ['freesurfer7'] and len(args) > 1\
                and args[0] in ['files', 'report', 'snapshot', 'tests']:
            ans = 1
            msg = 'Please confirm if you want FREESURFER7 or FREESURFER7_EXTRAS? '\
                  '(1) FREESURFER7 (2) FREESURFER7_EXTRAS ?'
            if not autorun:
                while ans not in ['1', '2']:
                    ans = input(msg)
            options = ['freesurfer7', 'freesurfer7extras']
            command = options[int(ans) - 1]
            parser = Command(command, args, x, destdir)
        elif command == 'spm12':
            ans = 3
            msg = 'Please confirm if you want SPM12_SEGMENT (T1-only)'\
                  ' or SPM12_SEGMENT_T1T2 (multimodal)? '\
                  ' (1) SPM12_SEGMENT (2) SPM12_SEGMENT_T1T2\n> '
            if not autorun:
                while ans not in ['1', '2']:
                    ans = input(msg)
                    rn, vn = 'SPM12_SEGMENT_T1T2', 'SPM12SegmentT1T2Validator'
                    if ans == '1':
                        rn, vn = 'SPM12_SEGMENT', 'SPM12SegmentValidator'
                    args.append((rn, vn))
            parser = Command('spm12', args, x, destdir)
        elif command == 'dartel':
            ans = 3
            msg = 'Please confirm if you want DARTEL_NORM2MNI (T1-only)' \
                  ' or DARTEL_NORM2MNI_T1T2 (multimodal)? ' \
                  ' (1) DARTEL_NORM2MNI (2) DARTEL_NORM2MNI_T1T2\n> '
            if not autorun:
                while ans not in ['1', '2']:
                    ans = input(msg)
                    rn, vn = 'DARTEL_NORM2MNI_T1T2', 'DartelNorm2MNIT1T2Validator'
                    if ans == '1':
                        rn, vn = 'DARTEL_NORM2MNI', 'DartelNorm2MNIValidator'
                    args.append((rn, vn))
            if args[0] == 'template' and len(args) == 2:
                args.insert(1, 'DARTEL_ALFA_20220301')
            parser = Command('dartel', args, x, destdir)
        elif command == 'dtifit':
            ans = 3
            msg = 'Please select the DWI protocol/sequence to retrieve results from:\n' \
                  '(1) DWI_ALFA1 (ALFA+, OPCIONAL,...)\n' \
                  '(2) DWI_MB2_PENSA (PENSA, BETA-AARC)\n> '
            if not autorun:
                while ans not in ['1', '2']:
                    ans = input(msg)
                    rn, vn = 'TOPUP_DTIFIT', 'DTIFITValidator'
                    if ans == '2':
                        rn, vn = 'DWIFSLPREPROC_DTIFIT', 'DWIFSLPreprocDTIFITValidator'
                    args.append((rn, vn))
            parser = Command('dtifit', args, x, destdir)

        try:
            parse(parser, commands[command])
        except ValueError as ve:
            log.error(ve)

    else:
        valid = [e for e, v in commands.items() if e != '']
        args = (command, '\n '.join(valid))
        msg = '%s not found \n\nValid commands:\n %s' % args
        log.error(msg)


class AParser(argparse.ArgumentParser):
    def _print_message(self, message, file=None):
        if message:
            print(message)


def create_parser():
    import bx
    cfgfile = op.join(op.expanduser('~'), '.xnat.cfg')
    commands = __find_all_commands__(bx)
    commands = {e.__name__.split('.')[-1].lower()[:-7]: e.__doc__
                for e in commands if e.__name__ != 'Command'}
    from bx import __version__
    logo = open(op.join(op.dirname(bx.__file__),
                        'data', 'logo'), encoding="utf8").read()
    desc = '[green on white]%s[/green on white]\n' % logo
    desc = desc + '[bold]bx[/bold] (v%s)\n\nAvailable commands:\n' % __version__

    for e, v in commands.items():
        i = int(len(str(e)) / 6)
        tabs = (3 - i) * '\t'
        v = '%s%s' % (tabs, v) if v is not None else ''
        desc = desc + ' [bold cyan]%s[/bold cyan] %s\n' % (e, str(v).split('\n')[0])

    desc = desc + '\nbx is distributed in the hope that it will be useful, ' \
                  'but WITHOUT ANY WARRANTY. \nSubmit issues/comments/PR at ' \
                  'http://gitlab.com/bbrc/xnat/bx.\n\nAuthors: Greg Operto, ' \
                  'Jordi Huguet, Marina Garcia Prat, Laura Ros, Ana Harris - ' \
                  'BarcelonaBeta Brain Research Center (2019-2025)'

    from argparse import RawTextHelpFormatter
    parser = AParser(description=desc, formatter_class=RawTextHelpFormatter)
    parser.add_argument('command', help='bx command')
    parser.add_argument('args', help='Arguments', nargs="*")
    parser.add_argument('--config', help='XNAT configuration file',
                        required=False, default=cfgfile)
    parser.add_argument('--dest', help='Destination folder',
                        required=False, action=ReadableDir)
    parser.add_argument('--verbose', '-v', action='store_true', default=False,
                        help='Display verbosal information (optional)',
                        required=False)

    from bx import __version__
    parser.add_argument('-V', '--version', action='version',
                        version="%(prog)s ("+__version__+")")

    return parser



from datetime import datetime
import os.path as op
import logging as log
import os


class Command(object):
    def __init__(self, command, args, xnat_instance, destdir):
        self.command = command
        self.args = args
        self.xnat = xnat_instance
        self.destdir = destdir

    def run_id(self, id, func, **kwargs):
        from bx import xnat

        cols = ['label', 'subject_label']
        if 'columns' in kwargs:
            cols = kwargs['columns']
            kwargs.pop('columns')

        is_max_rows = False
        if 'max_rows' in kwargs:
            max_rows = kwargs['max_rows']
            kwargs.pop('max_rows')
            is_max_rows = True
        if 'CI_TEST' not in os.environ.keys():
            max_rows = None
            is_max_rows = True
        if is_max_rows:
            experiments = xnat.collect_experiments(self.xnat, id, columns=cols,
                                                   max_rows=max_rows)
        else:
            experiments = xnat.collect_experiments(self.xnat, id, columns=cols)
        return func(self.xnat, experiments, **kwargs)

    def to_excel(self, df, index=True):
        dt = datetime.today().strftime('%Y%m%d_%H%M%S')
        args = ''
        if len(self.args) != 0:
            args = '_' + '_'.join(self.args)

        fn = 'bx_%s%s_%s.xlsx' % (self.command, args, dt)
        fp = op.join(self.destdir, fn)
        log.info('Saving it in %s' % fp)
        df.to_excel(fp, index=index)



from bx.command import Command
from bx import download as dl


class FMRIPrepCommand(Command):
    """fMRIPrep - Preprocessing of fMRI data.

    Available subcommands:
    files:\t\tdownload all `FMRIPREP` outputs (preprocessed BOLD, confounds, segmentations, everything...)
    report:\t\tdownload the validation report issued by `FMRIPrepValidator`
    tests:\t\tcreate an Excel table with all automatic tests outcomes from bbrc-validator

    Usage:
     bx fmriprep <subcommand> <resource_id>

    References:
    - Esteban O, et al., Nat Methods 16, 111–116 (2019).
    """
    nargs = 2
    resource_name = 'FMRIPREP'
    subcommands = ['files', 'report', 'tests']
    validator = 'FMRIPrepValidator'
    url = 'https://gitlab.com/bbrc/xnat/docker-images/-/tree/master/fmriprep'

    def __init__(self, *args, **kwargs):
        super(FMRIPrepCommand, self).__init__(*args, **kwargs)

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand in ['files', 'report']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand == 'tests':
            version = ['0f31a8ea', '*']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)



from bx.command import Command
from bx import download as dl
import os
import logging as log


class ASL3DCommand(Command):
    """Cerebral perfusion quantification of reconstructed 3D Arterial Spin Labeling MRI.

    Available subcommands:
     perfusion:\t\tcreates an Excel table with global perfusion values.
     aal:\t\tcreates an Excel table with regional perfusion values (in AAL atlas).
     maps:\t\tdownload the calibrated perfusion maps
     files:\t\tdownload all 3D-ASL QUANTIFICATION outputs (perfusion maps, files, everything...)
     snapshot:\t\tdownload a snapshot from the 3D-ASL QUANTIFICATION pipeline
     report:\t\tdownload the validation report issued by bbrc-validator
     tests:\t\tcreate an Excel table with all automatic tests outcomes from bbrc-validator

    Usage:
     bx asl3d <subcommand> <resource_id>

    References:
    - Chappell MA. et al., Imaging Neuroscience, 2023. DOI: 10.1162/imag_a_00041
    """
    nargs = 2
    resource_name = '3DASL_QUANTIFICATION'
    subcommands = ['perfusion', 'aal', 'maps', 'files', 'report', 'snapshot', 'tests']
    validator = 'ASL3DQuantificationValidator'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/asl#3d-asl-quantification'

    def __init__(self, *args, **kwargs):
        super(ASL3DCommand, self).__init__(*args, **kwargs)

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand in ['perfusion', 'aal']:
            sf = 'asl3d_%s' % subcommand
            df = self.run_id(id, dl.measurements, subfunc=sf,
                             resource_name=self.resource_name, max_rows=10)
            self.to_excel(df)

        elif subcommand == 'maps':
            debug = os.environ.get('CI_TEST', None)
            if debug:
                space = 'native_space'
            else:
                space = get_space()
            self.run_id(id, download_perfusion, resource_name=self.resource_name,
                        destdir=self.destdir, space=space, subcommand=subcommand)

        elif subcommand in ['files', 'report', 'snapshot']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand == 'tests':

            version = ['##7f43c4c4', '*']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)


def get_space():
    space = ''
    options = ['native_space', 'std_space', 'struct_space']
    msg = 'Which space?\n' \
          ' - native_space: native ASL space\n' \
          ' - std_space: MNI152 standard space \n' \
          ' - struct_space: structural T1 space\n'
    while space not in options:
        space = input(msg)
        if space not in options:
            m = 'Incorrect option (%s). Please try again.' % space
            print(m)
        else:
            space = space
    return space


def download_perfusion(x, experiments, resource_name, destdir, space, subcommand):
    import os.path as op
    from tqdm import tqdm
    from pyxnat.core.errors import DataError

    fp_options = {'native_space': 'masked_asl.nii.gz',
                  'std_space': 'asl2std.nii.gz',
                  'struct_space': 'asl2struct.nii.gz'}
    filename = fp_options[space]
    for e in tqdm(experiments):
        log.debug('Experiment %s:' % e['ID'])

        r = x.select.experiment(e['ID']).resource(resource_name)
        f = r.file(filename)

        fn = '%s_%s_%s_%s_%s.nii.gz' % (e['subject_label'], e['label'], e['ID'], 'perfusion', space)
        try:
            f.get(op.join(destdir, fn))
        except DataError as exc:
            log.error('Failed for %s. Skipping it. (%s)' % (e['ID'], exc))



from bx.command import Command
from bx import download as dl


class ASHSCommand(Command):
    """ASHS (Hippocampal subfield segmentation)

    Available subcommands:
     files:\t\tdownload all `ASHS` outputs (segmentation maps, volumes, everything...)
     volumes:\t\tcreates an Excel table with all hippocampal subfield volumes
     snapshot:\t\tdownload a snapshot from the `ASHS` pipeline
     report:\t\tdownload the validation report issued by `ASHSValidator`
     tests:\t\tcreates an Excel table with all automatic tests outcomes from `ASHSValidator`

    Usage:
     bx ashs <subcommand> <resource_id>
    """
    nargs = 2
    resource_name = 'ASHS'
    subcommands = ['volumes', 'files', 'report', 'snapshot', 'tests']
    validator = 'ASHSValidator'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/ashs'


    def __init__(self, *args, **kwargs):
        super(ASHSCommand, self).__init__(*args, **kwargs)

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]  # should be a project or an experiment_id
        if subcommand in ['volumes']:
            df = self.run_id(id, dl.measurements, subfunc=subcommand,
                             resource_name=self.resource_name)
            self.to_excel(df)

        elif subcommand in ['files', 'report', 'snapshot']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand == 'tests':
            version = ['*', 'beed8758', 'f80f2b13']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)



from bx.command import Command
from bx import download as dl
import os
import logging as log


class FdgCommand(Command):
    """18F-fluorodeoxyglucose PET imaging data

    Available subcommands:
     landau:\t\tcreates an Excel table with the Landau's metaROI signature
     maps:\t\tdownload the normalized FDG maps in MNI space
     files:\t\tdownload all outputs from the `FDG_QUANTIFICATION2` pipeline
     mri:\t\tcreates an Excel table with details from associated MRI sessions
     aging:\t\tcreates an Excel table with the aging composite ROI
     regional:\t\tcreates an Excel table with the regional quantification results
     snapshot:\t\tdownload snapshots from `FDG_QUANTIFICATION2` pipeline
     report:\t\tdownload the validation report issued by `FDGQuantificationValidator`
     tests:\t\tcollect all automatic tests outcomes from `FDGQuantificationValidator`

    Usage:
     bx fdg <subcommand> <resource_id>

    Reference:
    - Landau et al., Ann Neurol., 2012
    """
    nargs = 2
    subcommands = ['landau', 'maps', 'files', 'mri', 'aging', 'regional',
                   'snapshot', 'report', 'tests']
    resource_name = 'FDG_QUANTIFICATION2'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/pet'

    def __init__(self, *args, **kwargs):
        super(FdgCommand, self).__init__(*args, **kwargs)

    def parse(self):
        validator_name = 'FDGQuantificationValidator'

        subcommand = self.args[0]
        id = self.args[1]
        if subcommand == 'tests':
            version = ['*', '0390c55f']

            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=validator_name,
                             version=version, max_rows=25)
            self.to_excel(df)
        elif subcommand == 'landau':
            df = self.run_id(id, dl.measurements,
                             resource_name=self.resource_name,
                             subfunc=subcommand, max_rows=10)
            self.to_excel(df)
        elif subcommand == 'maps':
            fp = 'static_pet.nii.gz'
            if not os.environ.get('CI_TEST', None):
                fp = get_filepath(self.resource_name)
            self.run_id(id, download_pet, resource_name=self.resource_name,
                        filename=fp, destdir=self.destdir,
                        subcommand=subcommand)
        elif subcommand == 'mri':
            df = self.run_id(id, get_mri_table, validator='PetSessionValidator')
            self.to_excel(df)
        elif subcommand == 'aging':
            df = self.run_id(id, dl.measurements,
                             resource_name=self.resource_name,
                             subfunc=subcommand, max_rows=10)
            self.to_excel(df)
        elif subcommand == 'regional':
            atlas = 'hammers'
            if not os.environ.get('CI_TEST', None):
                atlas = get_atlas()
            df = self.run_id(id, dl.measurements,
                             resource_name=self.resource_name,
                             subfunc=f'pet_{atlas}', max_rows=10)
            self.args[0] += f'_{atlas}'
            self.to_excel(df)
        elif subcommand in ['files', 'report', 'snapshot']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=validator_name, destdir=self.destdir,
                        subcommand=subcommand)


class FtmCommand(Command):
    """18F-flutemetamol PET imaging data

    Available subcommands:
     centiloids:\tcreates an Excel table with centiloid scales
     maps:\t\tdownload the normalized FTM maps in MNI space
     files:\t\tdownload all outputs from the `FTM_QUANTIFICATION2` pipeline
     mri:\t\tcreates an Excel table with details from associated MRI sessions
     regional:\t\tcreates an Excel table with the regional quantification results
     snapshot:\t\tdownload snapshots from `FTM_QUANTIFICATION2` pipeline
     report:\t\tdownload the validation report issued by `FTMQuantificationValidator`
     tests:\t\tcollect all automatic tests outcomes from `FTMQuantificationValidator`

    Usage:
     bx ftm <subcommand> <resource_id>

    References:
     - Klunk et al, Alzheimers Dement., 2015
    """
    nargs = 2
    subcommands = ['centiloids', 'maps', 'files', 'mri', 'regional',
                   'snapshot', 'report', 'tests']
    resource_name = 'FTM_QUANTIFICATION2'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/pet'

    def __init__(self, *args, **kwargs):
        super(FtmCommand, self).__init__(*args, **kwargs)

    def parse(self):
        validator_name = 'FTMQuantificationValidator'

        subcommand = self.args[0]
        id = self.args[1]
        if subcommand == 'tests':
            version = ['*', '0390c55f']

            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=validator_name,
                             version=version, max_rows=25)
            self.to_excel(df)
        elif subcommand == 'centiloids':
            df = self.run_id(id, dl.measurements,
                             resource_name=self.resource_name,
                             subfunc=subcommand, max_rows=10)
            self.to_excel(df)
        elif subcommand == 'maps':
            fp = 'static_pet.nii.gz'
            if not os.environ.get('CI_TEST', None):
                fp = get_filepath(self.resource_name)
            self.run_id(id, download_pet, resource_name=self.resource_name,
                        filename=fp, destdir=self.destdir,
                        subcommand=subcommand)
        elif subcommand == 'mri':
            df = self.run_id(id, get_mri_table, validator='PetSessionValidator')
            self.to_excel(df)
        elif subcommand == 'regional':
            atlas = 'hammers'
            if not os.environ.get('CI_TEST', None):
                atlas = get_atlas()
            df = self.run_id(id, dl.measurements,
                             resource_name=self.resource_name,
                             subfunc=f'pet_{atlas}', max_rows=10)
            self.args[0] += f'_{atlas}'
            self.to_excel(df)
        elif subcommand in ['files', 'report', 'snapshot']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=validator_name, destdir=self.destdir,
                        subcommand=subcommand) 


class TauCommand(Command):
    """18F-RO-948 tau PET imaging data

    Available subcommands:
     tests:\t\tcollect all automatic tests outcomes from `TauPetSessionValidator`
     mri:\t\tcreates an Excel table with details from associated MRI sessions

    Usage:
     bx tau <subcommand> <resource_id>
    """
    nargs = 2
    subcommands = ['tests', 'mri']
    # resource_name = 'XXX'
    # url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/pet'

    def __init__(self, *args, **kwargs):
        super(TauCommand, self).__init__(*args, **kwargs)

    def parse(self):
        validator_name = 'TauPetSessionValidator'

        subcommand = self.args[0]
        id = self.args[1]
        if subcommand == 'tests':
            version = ['*', '0390c55f']

            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=validator_name,
                             version=version, max_rows=25)
            self.to_excel(df)
        elif subcommand == 'mri':
            df = self.run_id(id, get_mri_table, validator=validator_name)
            self.to_excel(df)


def get_filepath(resource_name, debug=False):
    region = None
    options = ['cgm', 'pons', 'wcbs', 'wc', 'wm', 'raw']
    msg = 'Reference region?\n'\
          ' - cgm: Cerebellar gray matter\n'\
          ' - wcbs: Whole cerebellum + Brain stem\n'\
          ' - pons: Pons\n'\
          ' - wc: Whole cerebellum\n'\
          ' - wm: White matter\n'\
          ' - raw: Raw image (not scaled, not normalized)\n> '
    if resource_name.startswith('FDG'):
        msg = msg.replace('Pons\n', 'Pons\n - vermis: Vermis\n')
        options.append('vermis')
    while region not in options and not debug:
        region = input(msg).lower()
        if region not in options:
            m = 'Incorrect option (%s). Please try again.' % region
            print(m)

    fp = ''
    if region == 'raw':
        fp = 'static_pet.nii.gz'
    else:
        fp = f'GLOBAL/wstatic_pet_scaled_{region}.nii.gz'

    return fp


def download_pet(x, experiments, resource_name, filename, destdir, subcommand):
    import os.path as op
    from tqdm import tqdm
    from pyxnat.core.errors import DataError

    for e in tqdm(experiments):
        log.debug('Experiment %s:' % e['ID'])

        r = x.select.experiment(e['ID']).resource(resource_name)
        f = r.file(filename)
        fn = '%s_%s_%s_%s' % (e['subject_label'], e['label'], e['ID'],
                              filename.replace('GLOBAL/', ''))
        try:
            f.get(op.join(destdir, fn))
        except DataError as exc:
            log.error('Failed for %s. Skipping it. (%s)' % (e, exc))


def get_mri_table(x, experiments, validator='PetSessionValidator'):
    import pandas as pd
    from tqdm import tqdm

    table = []
    for e in tqdm(experiments):
        try:
            e_id = e['ID']
            exp = x.array.experiments(experiment_id=e_id,
                                      columns=['date', 'subject_label']).data[0]
            if 'petSessionData' not in exp['xsiType']:
                continue

            row = [e_id, exp['date'], exp['project'], exp['subject_label']]

            r = x.select.experiment(e_id).resource('BBRC_VALIDATOR')
            if list(r.files('{}*'.format(validator))):
                has_t1 = r.tests(validator)['HasUsableT1']
                if has_t1['has_passed']:
                    row.extend(has_t1['data'][0].values())
            table.append(row)
        except KeyboardInterrupt:
            break

    cols = ['ID', 'date', 'project', 'subject_label',
            'MRI_ID', 'MRI_date', 'MRI_project', 'MRI_scanID', 'notes']
    df = pd.DataFrame(table, columns=cols).set_index('subject_label').sort_index()
    return df


def get_atlas(debug=False):
    atlas = None
    options = ['hammers', 'aal'] 
    msg = 'Regional quantification atlas?\n'\
          ' - hammers\n'\
          ' - aal\n> '
    while atlas not in options and not debug:
        atlas = input(msg).lower()
        if atlas not in options:
            m = 'Incorrect option (%s). Please try again.' % atlas
            print(m)
    return atlas



from bx.command import Command
from bx import download as dl
import os
import logging as log


class BASILCommand(Command):
    """BASIL - Bayesian Inference for Arterial Spin Labeling MRI.
    Arterial Spin Labeling (ASL) MRI is a non-invasive method for the quantification
    of perfusion.

    Available subcommands:
     perfusion:\t\tcreates an Excel table with global perfusion and arrival-time values.
     stats:\t\tcreates an Excel table with regional perfusion values (in Harvard-Oxford atlas).
     aal:\t\tcreates an Excel table with regional perfusion values (in AAL atlas).
     maps:\t\tdownload the calibrated perfusion maps
     files:\t\tdownload all BASIL-ASL outputs (perfusion maps, files, everything...)
     snapshot:\t\tdownload a snapshot from the BASIL-ASL pipeline
     report:\t\tdownload the validation report issued by bbrc-validator
     tests:\t\tcreate an Excel table with all automatic tests outcomes from bbrc-validator

    Usage:
     bx basil <subcommand> <resource_id>

    References:
    - Chappell MA., IEEE Transactions on Signal Processing, 2009.
    - Chappell MA. et al., Imaging Neuroscience, 2023. DOI: 10.1162/imag_a_00041
    """
    nargs = 2
    resource_name = 'BASIL'
    subcommands = ['perfusion', 'aal', 'stats', 'maps', 'files', 'report',
                   'snapshot', 'tests']
    validator = 'BASILValidator'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/asl#basil'

    def __init__(self, *args, **kwargs):
        super(BASILCommand, self).__init__(*args, **kwargs)

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand in ['perfusion', 'stats', 'aal']:
            sf = 'basil_%s' % subcommand
            df = self.run_id(id, dl.measurements, subfunc=sf,
                             resource_name='BASIL', max_rows=10)
            self.to_excel(df)

        elif subcommand == 'maps':
            debug = os.environ.get('CI_TEST', None)
            fp = get_filepath(debug=debug)
            self.run_id(id, download_perfusion, resource_name=self.resource_name,
                        filename=fp, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand in ['files', 'report', 'snapshot']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand == 'tests':

            version = ['##7f43c4c4', '*']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)


def get_filepath(debug=False):
    space = ''
    options = ['native_space', 'std_space', 'struct_space']
    msg = 'Which space?\n' \
          ' - native_space: native ASL space\n' \
          ' - std_space: MNI152 standard space \n' \
          ' - struct_space: structural T1 space\n'
    while not space in options and not debug:
        space = input(msg)
        if space not in options:
            m = 'Incorrect option (%s). Please try again.' % space
            print(m)
        else:
            space = space
    fp = '/{}/{}'.format(space, 'perfusion_calib.nii.gz')
    return fp


def download_perfusion(x, experiments, resource_name, filename, destdir, subcommand):

    import os.path as op
    from tqdm import tqdm
    from pyxnat.core.errors import DataError

    for e in tqdm(experiments):
        log.debug('Experiment %s:' % e['ID'])

        r = x.select.experiment(e['ID']).resource(resource_name)
        f = r.file(filename)
        space = filename.split('/')[1]
        fn = '%s_%s_%s_%s_%s.nii.gz' % (e['subject_label'], e['label'], e['ID'], 'perfusion', space)
        try:
            f.get(op.join(destdir, fn))
        except DataError as exc:
            log.error('Failed for %s. Skipping it. (%s)' % (e['ID'], exc))




commands = ['archiving tests',
            'ashs volumes',
            'ashs tests',
            'bamos volumes',
            'bamos stats',
            'bamos tests',
            'donsurf aparc',
            'donsurf tests',
            'dtifit tests',
            #'freesurfer6hires aparc',
            #'freesurfer6hires aseg',
            #'freesurfer6hires tests',
            #'freesurfer6hires hippoSfVolumes',
            'freesurfer7 aparc',
            'freesurfer7 aseg',
            'freesurfer7 tests',
            'freesurfer7 amygNucVolumes',
            'freesurfer7 hippoSfVolumes',
            'freesurfer7 hypothalamus',
            'freesurfer7 thalamus',
            'freesurfer7 brainstem',
            'fdg landau',
            'fdg aging',
            'fdg regional',
            'fdg tests',
            'fdg mri',
            'ftm centiloids',
            'ftm regional',
            'ftm tests',
            'ftm mri',
            'tau mri',
            'basil aal',
            'basil stats',
            'basil perfusion',
            'basil tests',
            'asl3d aal',
            'asl3d perfusion',
            'asl3d tests',
            'braak volumes',
            'braak thickness',
            'braak fdg',
            'signature jack',
            'signature dickerson',
            'scandates']


def dump(wd, config=None, interpreter='python', bash_command='/usr/bin/bash', debug=False):
    """ Call every command among those which return numeric endpoints
    (FreeSurfer thickness, volumes, hippocampal subfields, centiloids,
    signatures, Braak regions, etc). This results in a collection of
    spreadsheets to be served by dashboards."""
    import os
    import os.path as op
    import tempfile
    import bx

    alfa_projects = ['ALFA_PLUS_V1_20231120', 'ALFA_PLUS_V2_20230518', 'ALFA_20220301']
    fh, fp = tempfile.mkstemp(suffix='.sh')
    print(fp)
    os.close(fh)
    w = open(fp, 'w')
    cmd = 'mkdir %s' % op.join(wd, 'bx.new')
    w.write('export BX_DUMP=1\n')
    bx_fp = op.join(op.dirname(op.dirname(bx.__file__)), 'bin', 'bx')
    w.write(cmd + '\n')
    for c in commands:
        if 'fdg' in c:
            projects = ['ALFA_PET_FDG_20210421']
        elif 'ftm' in c:
            projects = ['ALFA_PET_FTM_20210421', 'ALFA_PLUS_V2_PET']
        elif 'tau' in c:
            projects = ['ALFA_PLUS_TAU']
        elif 'basil' in c:
            projects = ['ALFA_PLUS2']
        elif 'asl3d' in c:
            projects = ['ALFA_PLUS2_V2']
        else:
            projects = alfa_projects
        for p in projects:
            cmd = ' '.join([interpreter, bx_fp, c, p, '--dest', op.join(wd, 'bx.new')])
            if config:
                cmd += ' --config {}'.format(config)
            print(cmd)
            w.write(cmd + '\n')

    last_commands = ['rm -rf %s/bx.bak' % wd,
                     'mv %s/bx %s/bx.bak' % (wd, wd),
                     'mv %s/bx.new %s/bx' % (wd, wd)]
    for cmd in last_commands:
        w.write(cmd + '\n')

    w.close()

    if not debug:
        os.system('%s %s' % (bash_command, fp))

    # convert excel to csv
    from glob import glob
    import os.path as op
    import subprocess
    xlsx = glob(op.join(wd, 'bx', '*.xlsx'))
    for each in xlsx:
        fn = each.split('.xlsx')[0]
        call = [op.join(op.dirname(interpreter), 'xlsx2csv'), each, op.join(wd, 'bx', fn + '.csv')]
        print(' '.join(call))
        if not debug:
            subprocess.call(call)



import unittest
import os.path as op
import logging as log
import tempfile
import shutil
from pyxnat import Interface
from bx.parse import parse_args


def assert_excel(dd, values):
    import pandas as pd
    from glob import glob
    df = pd.read_excel(glob(op.join(dd, '*.xlsx'))[0], engine="openpyxl")
    df = df.values.tolist().pop()
    print(df)
    assert (df == values)


class RunThemAll(unittest.TestCase):

    def setUp(self):
        import bx
        from bx.parse import create_parser
        self.config_file = op.join(op.dirname(bx.__file__), '..', '.xnat.cfg')
        print(self.config_file)
        self.xnat_instance = Interface(config=self.config_file)
        create_parser()

    def test_check_xnat_item(self):
        from bx import xnat
        assert xnat.check_xnat_item(self.xnat_instance, 'testenv') == 0
        assert xnat.check_xnat_item(self.xnat_instance, 'BBRCDEV_E01613') == 1
        assert xnat.check_xnat_item(self.xnat_instance, 'DEV') == 2
        assert xnat.check_xnat_item(self.xnat_instance, 'foo') == -1

    def test_update_readme(self):
        import update_readme
        update_readme.update()

    def test_001_commands_without_args(self):
        """Command without args"""
        parse_args('scandates', [], self.xnat_instance)
        parse_args('freesurfer6hires', [], self.xnat_instance)
        parse_args('spm12', [], self.xnat_instance)
        parse_args('spm12', ['wrongsub'], self.xnat_instance)
        parse_args('spm12', ['wrongsub', 'testenv'], self.xnat_instance)
        parse_args('spm12', ['wrongsub', 'toto', 'testenv'], self.xnat_instance)

    def test_002_dates(self):
        """Extract imaging scan dates"""
        parse_args('scandates', ['testenv'], self.xnat_instance)
        parse_args('scandates', ['BBRCDEV_E01613'], self.xnat_instance)
        parse_args('scandates', ['toto'], self.xnat_instance)

    def test_003_freesurfer6_files(self):
        """FreeSurfer6 files"""
        dd = tempfile.mkdtemp()
        parse_args('freesurfer6', ['files', 'testenv'], self.xnat_instance,
                   destdir=dd)
        parse_args('freesurfer6', ['files', 'BBRCDEV_E01613'],
                   self.xnat_instance, destdir=dd)
        parse_args('freesurfer6', ['snapshot', 'BBRCDEV_E01613'],
                   self.xnat_instance, destdir=dd)
        parse_args('freesurfer6', ['report', 'BBRCDEV_E01613'],
                   self.xnat_instance, destdir=dd)

        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_004_freesurfer6_measurements(self):
        """FreeSurfer6 measurements"""
        dd = tempfile.mkdtemp()
        parse_args('freesurfer6', ['aparc', 'testenv'], self.xnat_instance, destdir=dd)
        print(dd)
        import pandas as pd
        from glob import glob
        df = pd.read_excel(glob(op.join(dd, '*.xlsx'))[0], engine="openpyxl")
        print(df)
        assert (len(df) > 1)
        eid = 'BBRCDEV_E00375'
        parse_args('freesurfer6', ['aseg', 'testenv'], self.xnat_instance, destdir=dd)
        for e in ['aparc', 'aseg', 'hippoSfVolumes']:
            parse_args('freesurfer6', [e, eid], self.xnat_instance, destdir=dd)
        parse_args('freesurfer6', ['tests', 'testenv'],
                   self.xnat_instance, destdir=dd)
        parse_args('freesurfer6', ['hippoSfVolumes', 'testenv'],
                   self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_005_spm12_project(self):
        """SPM12 project"""

        dd = tempfile.mkdtemp()
        parse_args('spm12', ['files', 'testenv'], self.xnat_instance, destdir=dd)
        parse_args('spm12', ['volumes', 'testenv'], self.xnat_instance,
                   destdir=dd)
        parse_args('spm12', ['tests', 'testenv'], self.xnat_instance, destdir=dd)
        parse_args('spm12', ['rc', 'testenv'], self.xnat_instance, destdir=dd)
        parse_args('spm12', ['rc', 'DEV'], self.xnat_instance, destdir=dd)

        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_006_spm12_experiment(self):
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E01613'
        for e in ['files', 'volumes', 'snapshot', 'rc']:
            parse_args('spm12', [e, eid], self.xnat_instance, destdir=dd)

        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_007_ashs_experiment(self):
        """ASHS experiment"""
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E02443'
        for e in ['files', 'volumes', 'tests', 'snapshot', 'report']:
            parse_args('ashs', [e, eid], self.xnat_instance, destdir=dd)

        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_008_dtifit(self):
        """DTIFIT"""
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E02801'
        for e in ['files', 'report', 'tests', 'snapshot', 'maps']:
            parse_args('dtifit', [e, eid], self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_009_id(self):
        """ID table"""
        dd = tempfile.mkdtemp()
        parse_args('id', ['BBRCDEV_E01613'], self.xnat_instance, destdir=dd)
        parse_args('id', ['testenv'], self.xnat_instance, destdir=dd)

        import pandas as pd
        from glob import glob
        df = pd.read_excel(glob(op.join(dd, '*.xlsx'))[0], engine="openpyxl")
        print(df)
        assert (len(df) > 1)
        shutil.rmtree(dd)

    def test_010_cat12_experiment(self):
        """CAT12 experiment"""
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E00375'
        parse_args('cat12', ['files', eid], self.xnat_instance, destdir=dd)
        parse_args('cat12', ['volumes', eid], self.xnat_instance, destdir=dd)

        values = [eid, 462026.0547849596, 350126.5098604396, 297766.4493393062,
                  'BASELINE_STUDIES']
        assert_excel(dd, values)

        parse_args('cat12', ['snapshot', eid], self.xnat_instance,
                   destdir=dd)
        parse_args('cat12', ['report', eid], self.xnat_instance,
                   destdir=dd)

        parse_args('cat12', ['rc', eid], self.xnat_instance,
                   destdir=dd)
        parse_args('cat12', ['rc', 'testenv'], self.xnat_instance,
                   destdir=dd)
        parse_args('cat12', ['rc', 'DEV'], self.xnat_instance,
                   destdir=dd)
        # parse_args('cat12', ['tests', 'testenv'], self.xnat_instance,
        # destdir=dd)

        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_011_nifti(self):
        """NIFTI"""
        dd1 = tempfile.mkdtemp()
        dd2 = tempfile.mkdtemp()
        eid = 'BBRCDEV_E02920'
        parse_args('nifti', [eid], self.xnat_instance, destdir=dd1)
        parse_args('nifti', ['all', eid], self.xnat_instance, destdir=dd2)
        from glob import glob
        uf = glob(op.join(dd1, '*.nii.gz'))
        af = glob(op.join(dd2, '*.nii.gz'))
        assert (len(uf) < len(af))
        assert (op.join(dd2, op.basename(uf[0])) in af)

        log.debug('Removing %s', dd1)
        shutil.rmtree(dd1)
        log.debug('Removing %s', dd2)
        shutil.rmtree(dd2)

    def test_cache(self):
        dd = tempfile.mkdtemp()
        from bx.cache import cache_freesurfer6, cache_ashs
        cache_freesurfer6(self.xnat_instance, 'testenv', od=dd, max_rows=10)
        cache_freesurfer6(self.xnat_instance, 'testenv', od=dd)
        cache_ashs(self.xnat_instance, 'testenv', od=dd)
        cache_ashs(self.xnat_instance, 'testenv', od=dd)
        shutil.rmtree(dd)

    def test_012_ants(self):
        """ANTS"""
        dd = tempfile.mkdtemp()
        for e in ['files', 'report', 'tests']:
            parse_args('ants', [e, 'BBRCDEV_E01613'], self.xnat_instance,
                       destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_013_archiving(self):
        """Archiving"""
        dd = tempfile.mkdtemp()
        parse_args('archiving', ['tests', 'BBRCDEV_E01613'], self.xnat_instance,
                   destdir=dd)
        parse_args('archiving', ['tests', 'testenv'], self.xnat_instance,
                   destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_014_lists(self):
        """lists"""
        parse_args('lists', ['display'], self.xnat_instance)

    def test_015a_fdg(self):
        """FDG"""
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E00745'
        for e in ['tests', 'landau', 'maps', 'files', 'mri', 'aging',
                  'regional', 'report', 'snapshot']:
            parse_args('fdg', [e, eid], self.xnat_instance, destdir=dd)
        parse_args('fdg', ['tests', 'testenv'],
                   self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_015b_ftm(self):
        """FTM"""
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E02124'
        for e in ['tests', 'centiloids', 'maps', 'files', 'mri',
                  'regional', 'report', 'snapshot']:
            parse_args('ftm', [e, eid], self.xnat_instance, destdir=dd)
        parse_args('ftm', ['tests', 'testenv'],
                   self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_016_signature(self):
        """signature"""
        dd = tempfile.mkdtemp()
        parse_args('signature', ['jack', 'BBRCDEV_E00375'],
                   self.xnat_instance, destdir=dd)
        parse_args('signature', ['dickerson', 'BBRCDEV_E02823'],
                   self.xnat_instance, destdir=dd)
        # parse_args('signature', [e, 'testenv'], self.xnat_instance,
        #           destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_018_jack_signature(self):
        """Jack's signature"""
        from bx import signature
        import bx

        fp = op.join(op.dirname(bx.__file__), '..', '.xnat.cfg')
        x = Interface(config=fp)

        eid = 'BBRCDEV_E00375'
        regions = ['entorhinal', 'inferiortemporal', 'middletemporal', 'fusiform']
        v = signature.__jack_signature__(x, eid, regions, weighted=True,
                                         measurement='ThickAvg', resource_name='FREESURFER6_HIRES')

        assert (v == 2.3719145266138897)

    def test_019_dickerson_signature(self):
        """Dickerson's signature"""
        from bx import signature
        import bx
        from math import isclose

        fp = op.join(op.dirname(bx.__file__), '..', '.xnat.cfg')
        x = Interface(config=fp)

        eid = 'BBRCDEV_E00375'
        d = {'ad': 2.0948601155910653,
             'aging': 2.1147848695063605}
        for sig, val in d.items():
            rois_path = op.join(op.dirname(bx.__file__), 'data',
                                'dickerson', sig)
            v = signature.__signature__(x, eid, rois_path, weighted=True,
                                        resource_name='FREESURFER6_HIRES')
            assert isclose(v, val, rel_tol=1e-05)

    def test_020_bamos(self):
        """BAMOS"""
        dd = tempfile.mkdtemp()
        for e in ['volumes', 'stats', 'files', 'layers']:
            parse_args('bamos', [e, 'BBRCDEV_E02832'],
                       self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_021_braak(self):
        """Braak regions"""
        dd = tempfile.mkdtemp()
        for e in ['volumes', 'thickness']:
            parse_args('braak', [e, 'BBRCDEV_E00375'], self.xnat_instance,
                       destdir=dd)
        parse_args('braak', ['fdg', 'BBRCDEV_E00745'], self.xnat_instance,
                   destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_022_donsurf(self):
        """DONSURF"""
        dd = tempfile.mkdtemp()
        parse_args('donsurf', ['aparc', 'BBRCDEV_E02824'], self.xnat_instance,
                   destdir=dd)
        from glob import glob
        import pandas as pd
        fp = glob(op.join(dd, '*.xlsx'))[0]
        print(fp)
        df = pd.read_excel(fp, engine='openpyxl')
        v = df.query('region == "bankssts" & side == "left"')['value'][0]
        assert (v == 8.288)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_023_freesurfer7(self):
        """FREESURFER7"""
        dd = tempfile.mkdtemp()
        subcommands = ['aparc', 'aseg', 'report', 'snapshot',
                       'hippoSfVolumes', 'tests', 'amygNucVolumes',
                       'brainstem',
                       'thalamus',
                       'hypothalamus']
        for e in subcommands:
            parse_args('freesurfer7', [e, 'BBRCDEV_E00270'], self.xnat_instance,
                       destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_024_spm12_t1t2(self):
        """MULTICHANNEL SPM12"""
        dd = tempfile.mkdtemp()
        subcommands = ['volumes', 'files', 'snapshot', 'report']
        for e in subcommands:
            v = ('SPM12_SEGMENT_T1T2', 'SPM12SegmentT1T2Validator')
            parse_args('spm12', [e, 'BBRCDEV_E02949', v], self.xnat_instance,
                       destdir=dd)
            v = ('SPM12_SEGMENT', 'SPM12SegmentValidator')
            parse_args('spm12', [e, 'BBRCDEV_E02949', v], self.xnat_instance,
                       destdir=dd)
        log.debug('Removing %s', dd)
        fp = op.join(dd, 'BBRCDEV_E02949', 'SPM12_SEGMENT_T1T2.zip')
        assert (op.isfile(fp))
        fp = op.join(dd, 'BBRCDEV_E02949', 'SPM12_SEGMENT.zip')
        assert (op.isfile(fp))
        shutil.rmtree(dd)

    def test_025_dump(self):
        from bx import dump
        dump.dump('/tmp', config=self.config_file, interpreter='python', debug=True)

    def test_026_basil(self):
        """BASIL-ASL"""
        dd = tempfile.mkdtemp()
        for e in ['perfusion', 'aal', 'stats', 'maps', 'files', 'report',
                  'snapshot', 'tests']:
            parse_args('basil', [e, 'BBRCDEV_E00276'], self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_027_freesurfer7extras(self):
        """FREESURFER7_EXTRAS"""
        dd = tempfile.mkdtemp()
        subcommands = ['report', 'snapshot', 'tests', 'files']
        for e in subcommands:
            parse_args('freesurfer7extras', [e, 'BBRCDEV_E02823'],
                       self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_028_tau(self):
        """TAU PET"""
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E02957'
        for e in ['tests', 'mri']:
            parse_args('tau', [e, eid], self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_029_dartel(self):
        """DARTEL"""
        dd = tempfile.mkdtemp()
        v1 = ('DARTEL_NORM2MNI', 'DartelNorm2MNIValidator')
        v2 = ('DARTEL_NORM2MNI_T1T2', 'DartelNorm2MNIT1T2Validator')
        for v in [v1, v2]:
            parse_args('dartel', ['template', 'DARTEL_ALFA_20220301', v],
                       self.xnat_instance, destdir=dd)
        subcommands = ['maps', 'report', 'tests']
        for e in subcommands:
            parse_args('dartel', [e, 'BBRCDEV_E02823', v1],
                       self.xnat_instance, destdir=dd)
            parse_args('dartel', [e, 'BBRCDEV_E02989', v2],
                       self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)

        items = ['Template_6_SPM12_SEGMENT.nii',
                 'Template_6_SPM12_SEGMENT_T1T2.nii',
                 'smwc1_10630_030410020719630_DARTEL_NORM2MNI.nii.gz',
                 'smwc1_12503_040512300518503_DARTEL_NORM2MNI_T1T2.nii.gz']
        for item in items:
            fp = op.join(dd, item)
            assert (op.isfile(fp))
        shutil.rmtree(dd)

    def test_030_qsmxt(self):
        """QSMxT"""
        dd = tempfile.mkdtemp()
        for e in ['stats', 'maps', 'files', 'report', 'tests']:
            parse_args('qsmxt', [e, 'BBRCDEV_E00276'], self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_031_mrtrix3(self):
        """MRtrix3"""
        dd = tempfile.mkdtemp()
        for e in ['connectome', 'report', 'snapshot', 'tests']:
            parse_args('mrtrix3', [e, 'BBRCDEV_E02885'], self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_032_xcpd(self):
        """XCP-D"""
        dd = tempfile.mkdtemp()
        for e in ['conmat', 'timeseries', 'report', 'snapshot', 'tests']:
            parse_args('xcpd', [e, 'BBRCDEV_E03088'], self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_033_fmriprep(self):
        """fMRIPrep"""
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E03098'
        for e in ['files', 'tests', 'report']:
            parse_args('fmriprep', [e, eid], self.xnat_instance, destdir=dd)

        log.debug('Removing %s', dd)
        shutil.rmtree(dd)
    
    def test_034_lcmodel(self):
        """LCModel"""
        dd = tempfile.mkdtemp()
        eid = 'BBRCDEV_E03082'
        for e in ['stats', 'correction', 'files', 'report',
                  'snapshot', 'tests']:
            parse_args('lcmodel', [e, eid], self.xnat_instance, destdir=dd)

        log.debug('Removing %s', dd)
        shutil.rmtree(dd)

    def test_035_lcmodel_stats_reformatting(self):
        """LCModel download_stats"""
        from bx.lcmodel import download_stats

        exps = [{"ID": "BBRCDEV_E00398"}, {"ID": "BBRCDEV_E03082"}]
        stats = download_stats(self.xnat_instance, exps)
        stats.sort_values(by=['region', 'subject'], inplace=True)

        assert stats.shape == (6, 72)
        assert list(stats['Cr']) == [7.917, 5.476, 4.544, 5.761, 8.093, 4.664]
        assert list(stats['SNR']) == [23.0, 28.0, 29.0, 33.0, 9.0, 9.0]
        q = 'id == "BBRCDEV_E00398" and region == "cuneus"'
        assert stats.query(q)['NAA+NAAG'].item() == 16.939

    def test_036_lcmodel_correction_merging(self):
        """LCModel download_correction"""
        from bx.lcmodel import download_correction

        exps = [{"ID": "BBRCDEV_E00398"}, {"ID": "BBRCDEV_E03082"}]
        corrs = download_correction(self.xnat_instance, exps)
        corrs.sort_values(by=['region', 'subject'], inplace=True)

        assert corrs.shape == (6, 11)
        assert list(corrs['gm']) == [44.64127726829608, 42.51328813576265,
                                     52.61325203056531, 43.43459982384211,
                                     49.97763657720883, 77.62208767982816]
        assert list(corrs['cov_gm']) == [0.4626002122020262, 0.4782981722285767,
                                         0.6379546884511307, 0.48617843354104,
                                         0.5137290744373951, 0.8719517243928585]
        q = 'id == "BBRCDEV_E00398" and region == "cuneus"'
        assert corrs.query(q)['corrected_csf'].item() == 1.119334437321481

    def test_037_asl3d(self):
        """BASIL-ASL"""
        dd = tempfile.mkdtemp()
        for e in ['perfusion', 'aal', 'maps', 'files',
                  'report', 'snapshot', 'tests']:
            parse_args('asl3d', [e, 'BBRCDEV_E03479'], self.xnat_instance, destdir=dd)
        log.debug('Removing %s', dd)
        shutil.rmtree(dd)


if __name__ == '__main__':
    unittest.main()



from bx.command import Command
from bx import download as dl
import logging as log


class XCPDCommand(Command):
    """XCP-D - Resting State fMRI postprocessing and functional connectivity.

    Available subcommands:
     conmat:\t\tdownload the functional connectivity matrix (Desikan-Killiany atlas)
     timeseries:\tdownload the parcellated BOLD time-series (Desikan-Killiany atlas)
     files:\t\tdownload all XCP_D outputs (denoised BOLD data, segmentations, everything...)
     report:\t\tdownload the validation report issued by `XCPDValidator`
     snapshot:\t\tdownload snapshots from the XCPD pipeline
     tests:\t\tcreate an Excel table with all automatic tests outcomes from bbrc-validator

    Usage:
     bx xcpd <subcommand> <resource_id>

    References:
    - Ciric R. et al., Nat Protoc 13 (2018).
    - Mehta K., Salo T. et al., bioRxiv (2023). DOI: 10.1101/2023.11.20.567926
    """
    nargs = 2
    resource_name = 'XCP_D'
    subcommands = ['conmat', 'timeseries', 'files', 'report', 'snapshot', 'tests']
    validator = 'XCPDValidator'
    url = 'https://gitlab.com/bbrc/xnat/docker-images/-/tree/master/xcp_d'

    def __init__(self, *args, **kwargs):
        super(XCPDCommand, self).__init__(*args, **kwargs)

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand == 'conmat':
            self.run_id(id, download_conmat, resource_name=self.resource_name,
                        destdir=self.destdir)
        elif subcommand == 'timeseries':
            self.run_id(id, download_ts, resource_name=self.resource_name,
                        destdir=self.destdir)
        elif subcommand in ['files', 'report', 'snapshot']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand == 'tests':
            version = ['f0eefa95', '*']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)


def download_conmat(x, experiments, resource_name, destdir, atlas='DK'):
    import os.path as op
    from tqdm import tqdm

    for e in tqdm(experiments):
        log.debug('Experiment %s:' % e['ID'])

        r = x.select.experiment(e['ID']).resource(resource_name)
        try:
            df = r.conmat(atlas)
        except StopIteration:
            log.error('Failed for %s. Skipping it.' % e['ID'])
            continue

        fn = '{}_{}_{}_xcpd_{}_conmat.xlsx'.format(e['subject_label'],
                                                   e['label'], e['ID'],
                                                   atlas)
        df.to_excel(op.join(destdir, fn))


def download_ts(x, experiments, resource_name, destdir, atlas='DK'):
    import os.path as op
    from tqdm import tqdm

    for e in tqdm(experiments):
        log.debug('Experiment %s:' % e['ID'])

        r = x.select.experiment(e['ID']).resource(resource_name)
        try:
            df = r.timeseries(atlas)
        except StopIteration:
            log.error('Failed for %s. Skipping it.' % e['ID'])
            continue

        fn = '{}_{}_{}_xcpd_{}_timeseries.xlsx'.format(e['subject_label'],
                                                       e['label'], e['ID'],
                                                       atlas)
        df.to_excel(op.join(destdir, fn), index=False)



import logging as log
import os


def scandate(x, experiment_id, series_desc='T1_ALFA1'):
    """ Given an experiment_id and a SeriesDescription, returns the acquisition
    date extracted from the first DICOM of the matching scanself.
    If multiple scans found, takes the one with the highest SeriesNumber."""

    columns = ['xsiType', 'xnat:imagescandata/type', 'xnat:imagescandata/ID']
    scans = x.array.scans(experiment_id=experiment_id, columns=columns).data
    t1_scans = {e['xnat:imagescandata/id']: e for e in scans
                if e['xnat:imagescandata/type'] == series_desc}

    if len(t1_scans.items()) == 0:
        msg = 'No T1 found for %s: %s. Trying with all of them.'\
              % (experiment_id, [e['xnat:imagescandata/id'] for e in scans])
        log.warning(msg)
        t1_scans = {e['xnat:imagescandata/id']: e for e in scans
                    if not e['xnat:imagescandata/id'].startswith('OT-')
                    and not e['xnat:imagescandata/id'].startswith('O-')}

    max_nb = sorted(t1_scans.keys())[-1]
    log.debug('Found scan: %s' % max_nb)

    scan = x.select.experiment(experiment_id).scan(max_nb)
    r = scan.resource('DICOM')

    return r.scandate()


def collect_scandates(x, experiments):

    def __create_table__(data):
        import pandas as pd
        columns = ['ID', 'label', 'subject_label', 'scandate', 'scantime']
        df = pd.DataFrame(data, columns=columns)
        #df['scandate'] = pd.to_datetime(df['scandate'])
        df = df.set_index('ID').sort_index()
        return df

    from tqdm import tqdm

    collected_data = []
    max_rows = 25 if os.environ.get('CI_TEST', None) else None

    for e in tqdm(experiments[:max_rows]):
        try:
            log.debug('Experiment ID: %s Subject label: %s'
                      % (e['ID'], e['subject_label']))

            if e['date'] == '':
                e['date'] = scandate(x, e['ID'])
            row = [e['ID'], e['label'], e['subject_label'], e['date'], e['time']]
            collected_data.append(row)

        except KeyboardInterrupt:
            return __create_table__(collected_data)
        except Exception as exc:
            log.error('Failed with %s. Skipping it. (%s)' % (e['ID'], exc))
            continue

    return __create_table__(collected_data)



from bx.command import Command
import tempfile


class ScanDatesCommand(Command):
    """Collect acquisition dates & times from imaging sessions.

    Usage:
     bx scandates <resource_id>
    """
    nargs = 1

    def __init__(self, *args, **kwargs):
        super(ScanDatesCommand, self).__init__(*args, **kwargs)

    def parse(self):
        id = self.args[0]  # should be a project or an experiment_id

        from bx.dcm import collect_scandates
        cols = ['label', 'subject_label',
                'xnat:experimentData/date', 'xnat:experimentData/time']
        res = self.run_id(id, collect_scandates, columns=cols, max_rows=3)
        if len(res) == 1:
            print(res.iloc[0])
        else:
            if self.destdir is None:
                self.destdir = tempfile.gettempdir()
            self.to_excel(res)



import logging as log
from tqdm import tqdm
import pandas as pd


def collect_reports(x, experiments, validator='ArchivingValidator',
                    version=['toto']):
    reports = {}

    for e in tqdm(experiments):
        try:
            exp = x.select.experiment(e['ID'])
            r = exp.resource('BBRC_VALIDATOR')
            if not r.exists():
                continue
            if len(list(r.files('%s*.json' % validator))) == 0:
                log.error('%s has no %s' % (e['ID'], validator))
                continue
            j = r.tests(validator)

            if 'version' not in j.keys():
                log.warning('Version not found in report %s' % j.keys())
                continue
            if j['version'] not in version and '*' not in version:
                continue
            fields = list(j.keys())
            try:
                for each in ['version', 'generated', 'experiment_id']:
                    fields.remove(each)
            except ValueError:
                msg = 'No valid report found (%s).' % e
                log.error(msg)
                raise Exception(msg)
            reports[e['ID']] = j

        except KeyboardInterrupt:
            return reports

    return reports


def validation_scores(x, experiments, validator, version):
    def build_dataframe(res):
        fields.insert(0, 'ID')
        for each in ['version', 'generated', 'experiment_id']:
            fields.remove(each)
        fields.extend(['version', 'generated'])
        df = pd.DataFrame(res, columns=fields).set_index('ID')
        return df

    res = []
    # Collecting reports of given version(s)
    log.info('Looking for experiments with %s report with versions %s.'
             % (validator, version))
    reports = dict(list(collect_reports(x, validator=validator,
                                        experiments=experiments,
                                        version=version).items()))
    if not reports:
        raise ValueError(f'No {validator} reports found, cannot create a table.')

    log.info('Now initiating download for %s experiment(s).'
             % len(reports.items()))

    # Creating list of columns (intersection between all reports tests)
    fields = list(list(reports.items())[0][1].keys())
    for e, report in tqdm(reports.items()):
        fields = set(fields).intersection(list(report.keys()))
    fields = list(fields)

    # Compiling data from reports (has_passed from each test)
    for e, report in tqdm(reports.items()):
        try:
            row = [e]
            row.extend([report[f]['has_passed'] for f in fields
                        if f not in ['version', 'generated', 'experiment_id']])
            row.extend([report[f] for f in ['version', 'generated']])
            res.append(row)
        except KeyboardInterrupt:
            return build_dataframe(res)

    # Building DataFrame
    return build_dataframe(res)



from bx.command import Command
from bx import download as dl
from bx.spm12 import SPM12Command

class DTIFITCommand(SPM12Command):
    """DTIFIT - Processing of Diffusion-weighted Imaging data

    Available subcommands:
     maps:\t\tdownload parametric maps (FA, MD, AxD, RD)
     files:\t\tdownload all `*DTIFIT` outputs (including parametric maps)
     report:\t\tdownload the validation report issued by `*DTIFITValidator`
     snapshot:\t\tdownload snapshots of the FA map, RGB tensor and TOPUP distortion correction map
     tests:\t\tcreate an Excel table with all automatic tests outcomes from bbrc-validator


    Usage:
     bx dtifit <subcommand> <resource_id>
    """
    nargs = 2
    resource_name = 'TOPUP_DTIFIT'
    subcommands = ['files', 'report', 'snapshot', 'tests', 'maps']
    validator = 'DTIFITValidator'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/dtifit'

    def __init__(self, *args, **kwargs):
        super(DTIFITCommand, self).__init__(*args, **kwargs)
        if self.resource_name == 'DWIFSLPREPROC_DTIFIT':
            self.url = 'https://gitlab.com/bbrc/xnat/docker-images/-/tree/master/dwifslpreproc_dtifit'

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand in ['files', 'report', 'snapshot', 'maps']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand == 'tests':
            version = ['*', '4e37c9d0']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)



import logging as log
from bx.spm12 import SPM12Command
from bx import download as dl


class DARTELCommand(SPM12Command):
    """DARTEL - Smoothed, spatially normalized and Jacobian scaled gray matter images in MNI space

    Available subcommands:
     maps:\t\tdownload the MNI-normalized gray matter maps
     template:\t\tdownload cohort-specific DARTEL template
     report:\t\tdownload the validation report issued by `DartelNorm2MNIValidator`
     tests:\t\tcreate an Excel table with all automatic tests outcomes from bbrc-validator

    Usage:
     bx dartel <subcommand> <resource_id>

    References:
     Ashburner J., Neuroimage, 2007. DOI: 10.1016/j.neuroimage.2007.07.007
    """
    nargs = 2
    resource_name = 'DARTEL_NORM2MNI'
    subcommands = ['maps', 'template', 'report', 'tests']
    validator = 'DartelNorm2MNIValidator'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/spm#dartel-norm2mni'

    def __init__(self, *args, **kwargs):
        super(DARTELCommand, self).__init__(*args, **kwargs)

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand in ['template']:
            download_template(self.xnat, template_label=id,
                              resource_name=self.resource_name,
                              destdir=self.destdir)

        elif subcommand in ['maps']:
            self.run_id(id, download_gm, resource_name=self.resource_name,
                        destdir=self.destdir)

        elif subcommand in ['report']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand == 'tests':
            version = ['*', '1229d1b7']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)


def download_gm(x, experiments, resource_name, destdir):
    import os.path as op
    from tqdm import tqdm

    filename = 'smw*'

    for e in tqdm(experiments):
        log.debug('Experiment {}:'.format(e['ID']))
        try:
            r = x.select.experiment(e['ID']).resource(resource_name)
            f = list(r.files(filename))[0]

            fn = 'smwc1_{}_{}_{}.nii.gz'.format(e['subject_label'],
                                                e['label'],
                                                resource_name)
            f.get(op.join(destdir, fn))
        except IndexError:
            log.error('Failed for {}. Skipping it.'.format(e['ID']))


def download_template(x, template_label, resource_name, destdir):
    import os.path as op

    filename = 'Template_6.nii'
    spm_res = 'SPM12_SEGMENT'
    if resource_name == 'DARTEL_NORM2MNI_T1T2':
        spm_res = 'SPM12_SEGMENT_T1T2'

    template_lib = x.select.project('DARTEL_TEMPLATES').subject(template_label)
    if not template_lib.exists():
        msg = '`{}` is not a template label'.format(template_label)
        raise ValueError(msg)
    log.info('List detected: {}'.format(template_label))

    f = template_lib.experiment(spm_res).resource('DARTEL').file(filename)
    fn = '{}_{}.nii'.format(op.splitext(filename)[0], spm_res)

    log.info('Saving template in {}'.format(op.join(destdir, fn)))
    f.get(op.join(destdir, fn))



from bx.command import Command
from bx import download as dl


class SPM12Command(Command):
    """SPM12 - Gray/white matter segmentation

    Available subcommands:
     files:\t\tdownload all `SPM12` outputs (segmentation maps, warp fields, everything...)
     volumes:\t\tcreates an Excel table with GM/WM/CSF volumes
     snapshot:\t\tdownload a snapshot from the segmentation results
     report:\t\tdownload the validation report issued by `SPM12Validator`
     tests:\t\tcreates an Excel table with all automatic tests outcomes from `SPM12`
     rc:\t\tdownloads rc* files (DARTEL imports)

    Usage:
     bx spm12 <subcommand> <resource_id>
    """
    nargs = 2
    resource_name = 'SPM12_SEGMENT'
    subcommands = ['volumes', 'files', 'report', 'snapshot', 'tests', 'rc']
    validator = 'SPM12SegmentValidator'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/spm'

    def __init__(self, *args, **kwargs):

        super(SPM12Command, self).__init__(*args, **kwargs)

        # SPM12Command is always provided with various arguments
        # We just check that the second (ie. the CLI actual args) has >1 item
        # The last one defines the type of segmentation (T1, T1T2) needed.
        last = args[1][-1] if len(args[1]) > 1 else []
        if isinstance(last, tuple) and len(last) == 2:
            (rn, vn) = last
            self.args.pop(-1)
            self.resource_name = rn
            self.validator = vn

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand in ['files', 'report', 'snapshot']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand in ['rc']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand='rc')

        elif subcommand == 'volumes':
            df = self.run_id(id, dl.measurements, subfunc=subcommand,
                             resource_name=self.resource_name, max_rows=10)
            self.to_excel(df)

        elif subcommand == 'tests':
            version = ['*', '0390c55f']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)



import logging as log
from bx.command import Command
from bx import download as dl


class LCModelCommand(Command):
    """LCModel - Quantification and tissue-correction of MR Spectroscopy images.

    Available subcommands:
     stats:\t\tcreates an Excel table with metabolite concentrations and global signal quality stats
     correction:\tcreates an Excel table with tissue correction parameters
     files:\t\tdownload all LCModel outputs (stats, tissue correction, masks, everything...)
     snapshot:\t\tdownload snapshots from the LCMODEL pipeline
     report:\t\tdownload the validation report issued by `LCModelValidator`
     tests:\t\tcreate an Excel table with all automatic tests outcomes from bbrc-validator

    Usage:
     bx lcmodel <subcommand> <resource_id>

    References:
    - Provencher SW., Magnetic Resonance in Medicine, 1993.
    """
    nargs = 2
    resource_name = 'LCMODEL'
    subcommands = ['stats', 'correction', 'files', 'report', 'snapshot', 'tests']
    validator = 'LCModelValidator'
    url = 'https://gitlab.com/bbrc/xnat/xnat-pipelines/-/tree/master/lcmodel'

    def __init__(self, *args, **kwargs):
        super(LCModelCommand, self).__init__(*args, **kwargs)

    def parse(self):
        subcommand = self.args[0]
        id = self.args[1]

        if subcommand == 'stats':
            df = self.run_id(id, download_stats)
            self.to_excel(df, index=False)

        if subcommand == 'correction':
            df = self.run_id(id, download_correction)
            self.to_excel(df, index=False)

        elif subcommand in ['files', 'report', 'snapshot']:
            self.run_id(id, dl.download, resource_name=self.resource_name,
                        validator=self.validator, destdir=self.destdir,
                        subcommand=subcommand)

        elif subcommand == 'tests':
            version = ['fd3dfdbd', '*']
            from bx import validation as val
            df = self.run_id(id, val.validation_scores,
                             validator=self.validator,
                             version=version, max_rows=25)
            self.to_excel(df)


def download_stats(x, experiments):
    """Download LCModel metabolite concentrations and other miscellaneous
    stats for a list of experiments and merge them in a single table"""
    import pandas as pd
    from tqdm import tqdm
    import io

    reshaped_df = pd.DataFrame()
    for e in tqdm(experiments):
        log.debug('Experiment %s:' % e['ID'])

        r = x.select.experiment(e['ID']).resource('LCMODEL')
        try:
            f = r.files('*lcmodel.xlsx')[0]
        except StopIteration:
            log.error('Failed for %s. Skipping it.' % e['ID'])
            continue

        with io.BytesIO(x.get(f._uri).content) as file_content:
            lcmodel_df = pd.read_excel(file_content)
        for reg in set(lcmodel_df.region):
            subj_region_df = lcmodel_df[lcmodel_df.region == reg]
            row_dict = {}
            for item in subj_region_df.to_dict('records'):
                for k in ['id', 'subject', 'region']:
                    row_dict[k] = item[k]
                if item['table'] in ['concentration', 'miscellaneous']:
                    row_dict[f"{item['measurement']}"] = item['value']
                    # row_dict[f"{item['measurement']}_units"] = item['units']
                elif item['table'] == 'concentration_SD':
                    row_dict[f"{item['measurement']}_SD"] = item['value']
                    # row_dict[f"{item['measurement']}_SD_units"] = item['units']
            row_df = pd.DataFrame(row_dict, index=[0])
            reshaped_df = pd.concat([reshaped_df, row_df])
    reshaped_df.reset_index(drop=True, inplace=True)
    return reshaped_df


def download_correction(x, experiments):
    """Download LCModel tissue correction values for a list of experiments
    and merge them in a single table"""
    import pandas as pd
    from tqdm import tqdm
    import io

    tissue_correction_df = pd.DataFrame()
    for e in tqdm(experiments):
        log.debug('Experiment %s:' % e['ID'])

        r = x.select.experiment(e['ID']).resource('LCMODEL')
        try:
            f = r.files('*mrs_tissue_corr.xlsx')[0]
        except StopIteration:
            log.error('Failed for %s. Skipping it.' % e['ID'])
            continue

        with io.BytesIO(x.get(f._uri).content) as file_content:
            df = pd.read_excel(file_content)
        tissue_correction_df = pd.concat([tissue_correction_df, df])
    tissue_correction_df.reset_index(drop=True, inplace=True)
    return tissue_correction_df




import logging as log
import os.path as op
import pandas as pd
import hashlib
from tqdm import tqdm
from datetime import datetime


def __find_cache__(od, md5, measurements=None):
    from glob import glob

    dates = sorted(set(['_'.join(op.split(e)[-1].split('_')[1:]).split('.')[0]\
        for e in glob(op.join(od, 'bxcache_%s'%md5, 'bxcache*.xlsx'))]))
    print(dates)
    if len(dates) == 0:
        return None
    lastdate = dates[-1]
    if not measurements is None:
        files = [op.join(od, 'bxcache_%s'%md5, 'bxcache_%s_%s.xlsx'%(n, lastdate)) \
            for n in measurements]
    else:
        files = [op.join(od, 'bxcache_%s'%md5, 'bxcache_%s.xlsx'%lastdate)]
    print(files)
    do_exist = True
    for f in files:
        if not op.isfile(f):
            do_exist = False
    if do_exist:
        log.info('Found cache %s'%lastdate)
        return files
    else:
        return None

def __save_files__(od, tables, md5, measurements=None):
    from datetime import datetime
    import os

    if not op.isdir(op.join(od, 'bxcache_%s'%md5)):
        os.mkdir(op.join(od, 'bxcache_%s'%md5))

    dt = datetime.today().strftime('%Y%m%d_%H%M%S')
    if not measurements is None:
        for n, each in zip(measurements, tables):

            fn = 'bxcache_%s_%s.xlsx'%(n, dt)
            fp = op.join(od, 'bxcache_%s'%md5, fn)
            log.info('Saving %s'%fp)
            each.to_excel(fp)

    else:
        fn = 'bxcache_%s.xlsx'%dt
        fp = op.join(od, 'bxcache_%s'%md5, fn)
        log.info('Saving %s'%fp)
        tables.to_excel(fp)

def cache_freesurfer6(c, project_id=None, resource_name='FREESURFER6',
        measurements=['aseg', 'aparc', 'hippoSfVolumes'], od='/tmp',
        force=False, max_rows=None):

    startTime = datetime.now()

    projects = list(set([project_id])) if isinstance(project_id, str) \
        else list(set(project_id))
    config = (projects, resource_name, measurements)
    md5 = hashlib.md5(str(config).encode('utf-8')).hexdigest()
    log.info('MD5: %s'%md5)

    cache = __find_cache__(od, md5, measurements)
    if cache and not force:
        log.info('Loading %s'%str(cache))
        res = [pd.read_excel(e, engine="openpyxl").set_index('subject') for e in cache]

        seconds = datetime.now() - startTime
        m, s = divmod(seconds.total_seconds(), 60)
        h, m = divmod(m, 60)
        elapsedtime = "%d:%02d:%02d" % (h, m, s)
        log.info('Elapsed time: %s'%elapsedtime)
        return res

    experiments = []
    for each in projects:
        a = c.array.experiments(project_id=each, columns=['subject_label'])
        experiments.extend(a.data)

    tables = {'aseg':[], 'aparc':[], 'hippoSfVolumes':[]}
    # Querying FreeSurfer resource -if any- from each experiment

    for e in tqdm(experiments[:max_rows]):
        s = e['subject_label']
        r = c.select.experiment(e['ID']).resource(resource_name)

        if r.exists():
            if 'aseg' in measurements:
                volumes = r.aseg()
                volumes['ID'] = e['ID']
                volumes['subject'] = s
                eTIV = volumes.query('region == "eTIV"')['value'].tolist()[0]
                volumes['eTIV'] = eTIV
                tables['aseg'].append(volumes)

            if 'aparc' in measurements:
                volumes = r.aparc()
                volumes['ID'] = e['ID']
                volumes['subject'] = s
                tables['aparc'].append(volumes)

            if 'hippoSfVolumes' in measurements:
                volumes = r.hippoSfVolumes()
                volumes['ID'] = e['ID']
                volumes['subject'] = s
                eTIV = r.aseg().query('region == "eTIV"')['value'].tolist()[0]
                volumes['eTIV'] = eTIV
                tables['hippoSfVolumes'].append(volumes)

    # Convert to dataframe
    df = [pd.concat(tables[each]).set_index('subject').sort_index() \
        for each in measurements]

    __save_files__(od, df, md5, measurements)

    seconds = datetime.now() - startTime
    m, s = divmod(seconds.total_seconds(), 60)
    h, m = divmod(m, 60)
    elapsedtime = "%d:%02d:%02d" % (h, m, s)
    log.info('Elapsed time: %s'%elapsedtime)
    return df


def cache_ashs(c, project_id=None, resource_name='ASHS',
         od='/tmp', force=False, max_rows=None):

    startTime = datetime.now()

    projects = list(set([project_id])) if isinstance(project_id, str) \
        else list(set(project_id))
    config = (projects, resource_name)
    md5 = hashlib.md5(str(config).encode('utf-8')).hexdigest()
    log.info('MD5: %s'%md5)

    cache = __find_cache__(od, md5)
    if cache and not force:
        log.info('Loading %s'%str(cache))
        res = [pd.read_excel(e, engine="openpyxl").set_index('subject') for e in cache]

        seconds = datetime.now() - startTime
        m, s = divmod(seconds.total_seconds(), 60)
        h, m = divmod(m, 60)
        elapsedtime = "%d:%02d:%02d" % (h, m, s)
        log.info('Elapsed time: %s'%elapsedtime)
        return res

    experiments = []
    for each in projects:
        a = c.array.experiments(project_id=each, columns=['subject_label'])
        experiments.extend(a.data)

    table = []
    # Querying ASHS resource -if any- from each experiment

    for e in tqdm(experiments[:max_rows]):
        s = e['subject_label']
        r = c.select.experiment(e['ID']).resource(resource_name)
        if r.exists():
            try:
                volumes = r.volumes()
                volumes['ID'] = e['ID']
                table.append(volumes)
            except IndexError:
                log.error('Error: check subject %s'%s)
                continue

    # Convert to dataframe
    df = pd.concat(table).set_index('subject').sort_index()
    __save_files__(od, df, md5)

    seconds = datetime.now() - startTime
    m, s = divmod(seconds.total_seconds(), 60)
    h, m = divmod(m, 60)
    elapsedtime = "%d:%02d:%02d" % (h, m, s)
    log.info('Elapsed time: %s'%elapsedtime)
    return df
